﻿using EdFi.Ods.Helpers;
using RestSharp;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Net;

namespace EdFi.Ods.Labs.Tests
{
    public static class SdkHelper
    {
        public static RestClient CreateClient(int? schoolYear = null)
        {
            // Trust all SSL certs -- needed for now until signed SSL certificates are configured.
            ServicePointManager.ServerCertificateValidationCallback = ((sender, certificate, chain, sslPolicyErrors) => true);

            // Oauth configuration
            var odsBaseUri = ConfigurationManager.AppSettings["EdFiOdsBaseUri"];
            var odsApiUri = ConfigurationManager.AppSettings["EdFiOdsApiUri"];
            var odsKey = ConfigurationManager.AppSettings["EdFiOdsKey"];
            var odsSecret = ConfigurationManager.AppSettings["EdFiOdsSecret"];

            // TokenRetriever makes the oauth  calls
            var tokenRetriever = new TokenRetriever(odsBaseUri, odsKey, odsSecret);
            var client = schoolYear != null ?
                new RestClient(odsApiUri + schoolYear + "/")
                : new RestClient(odsApiUri);

            client.Authenticator = new BearerTokenAuthenticator(tokenRetriever);

            return client;
        }

        public static void UploadInterchanges(RestClient client, List<Interchange> interchanges)
        {
            var bulkOperationsApi = new EdFi.Rest.Apis.BulkOperationsApi(client);
            var postResponse =
                bulkOperationsApi.PostBulkOperations(new EdFi.Rest.Models.BulkOperation
                {
                    uploadFiles = interchanges.Select(x => new EdFi.Rest.Models.UploadFile
                    {
                        format = x.Format,
                        interchangeType = x.InterchangeType,
                        size = x.Length
                    }).ToList()
                });

            var bulkOperation = postResponse.Data;
            var uploadApi = new EdFi.Rest.Apis.UploadsApi(client);

            interchanges.ForEach(x => x.BulkOperationId = bulkOperation.id);

            interchanges.ForEach(x =>
                x.FileUploadId = bulkOperation.uploadFiles.Single(uploadFile =>
                    uploadFile.interchangeType.ToLower() == x.InterchangeType.ToLower() && uploadFile.size == x.Length).id);

            foreach (var interchange in interchanges)
            {
                using (var stream = interchange.CreateStream())
                {
                    ChunkUpload(uploadApi, interchange.FileUploadId, stream, interchange.Length);
                }
            }

            bulkOperation = bulkOperationsApi.GetBulkOperationsById(bulkOperation.id).Data;

            // Commit the files.
            bulkOperation.uploadFiles.ForEach(x => uploadApi.CommitUploads(new EdFi.Rest.Models.Upload { id = x.id }));

            // Not necessarily recommended for the real world, but this is polling by the second until completion.
            while (bulkOperation.status != "Completed" && bulkOperation.status != "Error")
            {
                System.Threading.Thread.Sleep(1000);
                bulkOperation = bulkOperationsApi.GetBulkOperationsById(bulkOperation.id).Data;
            }

            // While we're here, we could do additional work around retrieving exceptions.
            var bulkOperationExceptionsApi = new EdFi.Rest.Apis.BulkOperationExceptionsApi(client);
            string exceptionContent = string.Empty;

            foreach (var uploadFile in bulkOperation.uploadFiles)
            {
                var offset = 0;
                var limit = 50;
                IRestResponse<List<EdFi.Rest.Models.BulkOperationException>> exceptions = null;

                do
                {
                    exceptions = bulkOperationExceptionsApi.GetBulkOperationExceptions(bulkOperation.id, uploadFile.id, limit, offset);

                    var exceptionContentThisFile = exceptions.Content;
                    
                    foreach (var bulkOperationException in exceptions.Data)
                    {
                        System.Diagnostics.Trace.WriteLine("Exception Message: '{0}'", bulkOperationException.message);
                    }

                    if (exceptionContentThisFile != null) exceptions.Content += " ";

                    exceptionContent += exceptionContentThisFile;

                    var interchange = interchanges.Single(x => x.FileUploadId == uploadFile.id);

                    interchange.Exceptions.AddRange(exceptions.Data);

                    offset += limit;
                } while (exceptions.Data.Count > 0);
            }
        }

        private static void ChunkUpload(EdFi.Rest.Apis.UploadsApi uploadApi, string uploadId, Stream stream, long fileLength)
        {
            int offset = 0;
            const long maximumChunkSize = 4000000;

            // Loop to simulate chunking.
            while (offset < fileLength)
            {
                var chunkSize = (offset + maximumChunkSize) <= fileLength
                    ? maximumChunkSize
                    : fileLength - offset;
                var fileBytes = new byte[chunkSize];

                stream.Read(fileBytes, 0, fileBytes.Length);

                var uploadResponse = uploadApi.PostUploads(new EdFi.Rest.Models.Upload
                {
                    id = uploadId,
                    size = fileBytes.Length,
                    offset = offset,
                    // These will be file bytes for the real implementation.
                    fileBytes = fileBytes
                });

                offset += (int)chunkSize;
            }
        }
    }
}
